/*
 *  CUPS backend for a named pipe.
 *  Copyright CANON INC. 2004-2009
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
//
//  '2004.05.25
//  Using the original "usb.c" code of the CUPS 1.1.14 to create a new
//  backend for writing the printing data into a named pipe.
//

/*
 * "$Id: ccp.c,v 1.1.1.1 2008/12/18 06:29:04 arai_norihiro Exp $"
 *
 *   USB port backend for the Common UNIX Printing System (CUPS).
 *
 *   Copyright 1997-2002 by Easy Software Products, all rights reserved.
 *
 *   These coded instructions, statements, and computer programs are the
 *   property of Easy Software Products and are protected by Federal
 *   copyright law.  Distribution and use rights are outlined in the file
 *   "LICENSE" which should have been included with this file.  If this
 *   file is missing or damaged please contact Easy Software Products
 *   at:
 *
 *       Attn: CUPS Licensing Information
 *       Easy Software Products
 *       44141 Airport View Drive, Suite 204
 *       Hollywood, Maryland 20636-3111 USA
 *
 *       Voice: (301) 373-9603
 *       EMail: cups-info@cups.org
 *         WWW: http://www.cups.org
 *
 * Contents:
 *
 *   main()         - Send a file to the specified USB port.
 *   list_devices() - List all USB devices.
 */




#include <cups/cups.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <unistd.h>
#include <fcntl.h>
#include <limits.h>




#define	CPRTD_FIFO_PATH		"/var/ccpd/fifo%d"
#define	CPRTD_BACKEND_NAME	"ccp"
#define	CPRTD_MAX_FIFO		16


#define	DATA_BUF_SIZE		(PIPE_BUF)


int	g_canceled = 0;

FILE *g_log_fp;


void list_devices(void)
{
	char device[256];
	int i;

	for( i = 0 ; i < CPRTD_MAX_FIFO ; i++ )
	{
		sprintf(device, CPRTD_FIFO_PATH, i);
		if( !access(device, 0) )
		{
      		printf(
			"direct %s:%s \"Unknown\" \"Canon Priter Daemon Port#%d\"\n",
			CPRTD_BACKEND_NAME, device, i + 1);
		}
    }
}

void make_header(unsigned int job, unsigned int bytes, char buf[8])
{
	buf[0] = job & 0xff;
	buf[1] = (job >> 8)  & 0xff;
	buf[2] = (job >> 16) & 0xff;
	buf[3] = (job >> 24) & 0xff;

	buf[4] = bytes & 0xff;
	buf[5] = (bytes >> 8)  & 0xff;
	buf[6] = (bytes >> 16) & 0xff;
	buf[7] = (bytes >> 24) & 0xff;
}

int write_data(int ofd, unsigned char *data, int data_bytes)
{
	int written_bytes = 0;
	int bytes;
	unsigned char *p_data = data;

#ifdef	DEBUG
	{
		int size
			=  (unsigned int)data[4]
			+ ((unsigned int)data[5] << 8)
			+ ((unsigned int)data[6] << 16)
			+ ((unsigned int)data[7] << 24);
		fprintf(g_log_fp, "DEBUG: ccp:wrote %d bytes, size=%d\n",
		data_bytes, size);
	}
#endif

	while( !g_canceled && data_bytes > 0 )
	{
		bytes = write(ofd, p_data, data_bytes);

		if( bytes < 0 )
		{
			if( errno == EINTR )
				continue;

			fprintf(g_log_fp, "ERROR: %s write error,%d\n",
					CPRTD_BACKEND_NAME, errno);

			goto error_exit;
		}
		data_bytes -= bytes;
		written_bytes += bytes;
		p_data += bytes;
	}

	return written_bytes;

error_exit:
	fprintf(g_log_fp, "ERROR: ccp: write_data error, exit\n");
	return -1;
}

int send_data(int ifd, int ofd, int job)
{
	char data_buf[DATA_BUF_SIZE];

	while( !g_canceled )
	{
		int read_bytes = read(ifd, data_buf + 8, DATA_BUF_SIZE - 8);

		if( read_bytes > 0 )
		{
			int write_bytes = read_bytes + 8;

			make_header(job, read_bytes, data_buf);

			if( write_data(ofd, data_buf, write_bytes) < write_bytes )
				goto error_exit;
		}
		else if( read_bytes < 0 )
		{
			if( errno == EINTR )
				continue;
			fprintf(g_log_fp, "ERROR: %s read error,%d\n",
			CPRTD_BACKEND_NAME, errno);
			goto error_exit;
		}
		else
		{
			break;
		}
	}

	make_header(job, 0, data_buf);

	fprintf(g_log_fp, "DEBUG: ccp: last data.\n");

	if( write_data(ofd, data_buf, 8) < 8 )
		goto error_exit;

	fprintf(g_log_fp, "DEBUG: ccp: end of send data.\n");
	return 0;

error_exit:
	fprintf(g_log_fp, "ERROR: ccp: send_data error, exit\n");
	return -1;
}

void signal_handler(int sigcode)
{
	g_canceled = 1;
}

int main(int argc, char *argv[])
{
	struct sigaction sigact;
	char resource[1024];
	char *options;
	int job;
	int copies;
	int ifd = 0, ofd;
	int ret;

#ifdef DEBUG_SLEEP
	sleep(30);
#endif

	setbuf(stderr, NULL);

#ifdef	DEBUG_LOG
	g_log_fp = fopen("/tmp/ccp_log.txt", "w");
#else
	g_log_fp = stderr;
#endif

	if( argc == 1 )
	{
		list_devices();
		return 0;
	}
	else if( argc < 6 || argc > 7 )
	{
		fprintf(stderr, "Usage: %s job-id user title copies options [file]\n",
		CPRTD_BACKEND_NAME);
		return 1;
	}

	memset(&sigact, 0, sizeof(sigact));
	sigact.sa_handler = signal_handler;
	if( sigaction(SIGTERM, &sigact, NULL) )
	{
		return 1;
	}

	memset(&sigact, 0, sizeof(sigact));
	sigact.sa_handler = signal_handler;
	if( sigaction(SIGPIPE, &sigact, NULL) )
	{
		return 1;
	}

	if( argc == 7 )
	{
		if( (ifd = open(argv[6], O_RDONLY)) == -1 )
		{
			perror("ERROR: Can't open file");
			return 1;
		}
		copies = atoi(argv[4]);
	}

	if( (strncmp(argv[0], "ccp:/var/ccpd", 13) == 0) )
	{
		memset(resource, 0, 1024);
		strncpy(resource, argv[0] + 4, 1023);
	}
	else
	{
		perror("ERORR: Illegal backend");
		return (1);
	}

	if( (options = strchr(resource, '?')) != NULL )
	{
		*options++ = '\0';
	}

	if( (ofd = open(resource, O_WRONLY | O_EXCL)) == -1 )
	{
		perror("ERROR: Can't open FIFO");
		return 1;
	}

	job = atoi(argv[1]);

	if( send_data(ifd, ofd, job) < 0 )
		ret = 1;
	else
		ret = 0;

	if( ifd != 0 )
		close(ifd);

	close(ofd);

#ifdef	DEBUG_LOG
	if( g_log_fp != NULL )
		fclose(g_log_fp);
#endif

	return ret;
}

